package jdk.management.resource.internal.inst;

import java.net.SocketOptions;
import sun.misc.JavaIOFileDescriptorAccess;
import java.io.FileDescriptor;
import sun.misc.SharedSecrets;
import jdk.internal.instrumentation.InstrumentationMethod;
import jdk.management.resource.ResourceRequest;
import jdk.management.resource.ResourceRequestDeniedException;
import java.io.IOException;
import jdk.management.resource.internal.ApproverGroup;
import jdk.management.resource.internal.ResourceIdImpl;
import java.net.SocketAddress;
import java.net.InetAddress;
import jdk.internal.instrumentation.TypeMapping;
import jdk.internal.instrumentation.InstrumentationTarget;

@InstrumentationTarget("java.net.Socket")
@TypeMapping(from = "jdk.management.resource.internal.inst.SocketRMHooks$SocketImpl", to = "java.net.SocketImpl")
public final class SocketRMHooks
{
    private boolean created;
    SocketRMHooks.SocketImpl impl;
    
    public SocketRMHooks() {
        this.created = false;
    }
    
    public InetAddress getLocalAddress() {
        return this.getLocalAddress();
    }
    
    @InstrumentationMethod
    public void bind(SocketAddress socketAddress) throws IOException {
        ResourceIdImpl of = null;
        ResourceRequest approver = null;
        long request = 0L;
        if (!this.isBound()) {
            of = ResourceIdImpl.of(socketAddress);
            approver = ApproverGroup.SOCKET_OPEN_GROUP.getApprover(this);
            try {
                request = approver.request(1L, of);
                if (request < 1L) {
                    throw new IOException("Resource limited: too many open sockets");
                }
            }
            catch (ResourceRequestDeniedException cause) {
                throw new IOException("Resource limited: too many open sockets", cause);
            }
        }
        int n = 0;
        try {
            this.bind(socketAddress);
            n = 1;
        }
        finally {
            if (approver != null) {
                approver.request(-(request - (long)n), of);
            }
        }
    }
    
    @InstrumentationMethod
    public boolean isBound() {
        return this.isBound();
    }
    
    @InstrumentationMethod
    public void connect(SocketAddress socketAddress, int n) throws IOException {
        ResourceIdImpl of = null;
        ResourceRequest approver = null;
        long request = 0L;
        if (!this.isBound()) {
            of = ResourceIdImpl.of(this.getLocalAddress());
            approver = ApproverGroup.SOCKET_OPEN_GROUP.getApprover(this);
            try {
                request = approver.request(1L, of);
                if (request < 1L) {
                    throw new IOException("Resource limited: too many open sockets");
                }
            }
            catch (ResourceRequestDeniedException cause) {
                throw new IOException("Resource limited: too many open sockets", cause);
            }
        }
        int n2 = 0;
        try {
            this.connect(socketAddress, n);
            n2 = 1;
        }
        finally {
            if (approver != null) {
                approver.request(-(request - (long)n2), of);
            }
        }
    }
    
    @InstrumentationMethod
    final void postAccept() {
        this.postAccept();
        FileDescriptor fileDescriptor = this.impl.getFileDescriptor();
        JavaIOFileDescriptorAccess javaIOFileDescriptorAccess = SharedSecrets.getJavaIOFileDescriptorAccess();
        long handle;
        try {
            handle = javaIOFileDescriptorAccess.getHandle(fileDescriptor);
            if (handle == -1L) {
                handle = (long)javaIOFileDescriptorAccess.get(fileDescriptor);
            }
        }
        catch (UnsupportedOperationException ex) {
            handle = (long)javaIOFileDescriptorAccess.get(fileDescriptor);
        }
        ResourceIdImpl of = ResourceIdImpl.of(Long.valueOf(handle));
        ResourceRequest approver = ApproverGroup.FILEDESCRIPTOR_OPEN_GROUP.getApprover(fileDescriptor);
        long request = 0L;
        boolean n = false;
        try {
            request = approver.request(1L, of);
            if (request < 1L) {
                throw new ResourceRequestDeniedException("Resource limited: too many open file descriptors");
            }
            n = true;
        }
        finally {
            if (n == 0) {
                try {
                    this.close();
                }
                catch (IOException ex2) {}
                approver.request(-Math.max(0L, request - 1L), of);
            }
        }
    }
    
    @InstrumentationMethod
    public boolean isClosed() {
        return this.isClosed();
    }
    
    @InstrumentationMethod
    public synchronized void close() throws IOException {
        if (this.isClosed()) {
            return;
        }
        int bound = this.isBound();
        InetAddress localAddress = this.getLocalAddress();
        try {
            this.close();
        }
        finally {
            if (!(bound == 0)) {
                ApproverGroup.SOCKET_OPEN_GROUP.getApprover(this).request(-1L, ResourceIdImpl.of(localAddress));
            }
        }
        return;
    }
    
    abstract class SocketImpl implements SocketOptions
    {
        protected FileDescriptor fd;
        
        protected FileDescriptor getFileDescriptor() {
            return this.fd;
        }
    }
}
